 

// The wiring for 2 TFT displays to an ESP32 is described in the
// "wiring" tab of this sketch.

// Configuration settings for the eye, eye style, display count,
// chip selects and x offsets can be defined in the sketch "config.h" tab.

// Performance (frames per second = fps) can be improved by using
// DMA (for SPI displays only) on ESP32 and STM32 processors. Use
// as high a SPI clock rate as is supported by the display. 27MHz
// minimum, some displays can be operated at higher clock rates in
// the range 40-80MHz.

// Single defaultEye performance for different processors
//                                  No DMA   With DMA
// ESP8266 (160MHz CPU) 40MHz SPI   36 fps
// ESP32 27MHz SPI                  53 fps     85 fps
// ESP32 40MHz SPI                  67 fps    102 fps
// ESP32 80MHz SPI                  82 fps    116 fps // Note: Few displays work reliably at 80MHz
// STM32F401 55MHz SPI              44 fps     90 fps
// STM32F446 55MHz SPI              83 fps    155 fps
// STM32F767 55MHz SPI             136 fps    197 fps

// DMA can be used with RP2040, STM32 and ESP32 processors when the interface
// is SPI, uncomment the next line:
// Load TFT driver library
// This struct is populated in config.h
typedef struct {        // Struct is defined before including config.h --
	int8_t  select;       // pin numbers for each eye's screen select line
	int8_t  wink;         // and wink button (or -1 if none) specified there,
	float  xang, yang;     // also display rotation and the x offset
} eyeInfo_t;

#define GC9A01A_SLPIN 0x10     ///< Enter Sleep Mode
#define GC9A01A_SLPOUT 0x11    ///< Sleep Out

TaskHandle_t Task1;
#define GC9A01_DRIVER
#define NUM_EYES 1
#define USE_DMA
//#include <SPI.h>
#include <TFT_eSPI.h>

//now override the originals

#define TFT_MOSI  23//PIN_D7  // Automatically assigned with ESP8266 if not defined
#define TFT_SCLK  18//PIN_D5  // Automatically assigned with ESP8266 if not defined




       // A single instance is used for 1 or 2 displays
// A pixel buffer is used during eye rendering
//#define BUFFER_SIZE 256 // 128 to 1024 seems optimum
#define SCREEN_WIDTH  240
#define SCREEN_HEIGHT 240
uint16_t fullscreen = SCREEN_HEIGHT * SCREEN_WIDTH;
uint32_t Center = SCREEN_WIDTH / 2 * SCREEN_WIDTH + SCREEN_HEIGHT / 2;
uint32_t icen = 70 * 140 + 70;



uint16_t *MainScreenTop; //screen buffer
uint16_t *MainScreenBottom; //screen buffer
uint8_t nScreen = 0;
#define SCREEN_HALF 120
#define SCR_HALF_SQ  14400 // 120 * 120
bool DoFrame = false;

#include "config.h"     // ****** CONFIGURATION IS DONE IN HERE ******

#define SCREEN_X_START 0
#define SCREEN_X_END   SCREEN_WIDTH   // Badly named, actually the "eye" width!
#define SCREEN_Y_START 0
#define SCREEN_Y_END   SCREEN_HEIGHT // Actually "eye" height

struct {                // One-per-eye structure
	int16_t   tft_cs;     // Chip select pin for each display
	int16_t xang, yang;
	int16_t irissize;

} eye[NUM_EYES];

uint32_t startTime;  // For FPS indicator
TFT_eSPI tft;

// INITIALIZATION -- runs once at startup ----------------------------------
void setup(void) {
	Serial.begin(115200);
	//while (!Serial);
	//Serial.print("Starting on core :");
	//Serial.println(xPortGetCoreID());
	// User call for additional features
	Serial.print("Free Heap:");
	Serial.println(ESP.getFreeHeap());
	Serial.print("Min Free Heap:");
	Serial.println(ESP.getMinFreeHeap());
	Serial.print("Free HeapSize:");
	Serial.println(ESP.getHeapSize());
	Serial.print("MaxAlloc Heap:");
	Serial.println(	ESP.getMaxAllocHeap());
	 

	// Initialise the eye(s), this will set all chip selects low for the tft.init()
	initEyes();
	
	// Initialise TFT
	Serial.println("Initialising displays");
	tft.init();
	 

	Serial.println("Initialising Config");

	MainScreenTop = (uint16_t*)malloc(SCREEN_HEIGHT * SCREEN_WIDTH );
	if (MainScreenTop == NULL)
		Serial.println("Top Screen Failed");
	else
		Serial.println("Top Screen Created");
	MainScreenBottom = (uint16_t*)malloc(SCREEN_HEIGHT * SCREEN_WIDTH );
	if (MainScreenTop == NULL)
		Serial.println("Bottom Screen Failed");
	else
		Serial.println("Bottom Screen Created");

	 

	Serial.println("Initialising DMA");
#ifdef USE_DMA
	tft.initDMA();
#endif


	Serial.println("Screen Done");
	 

	eye[0].xang = 0;
	eye[0].yang = 0;
	eye[0].irissize = 5;
	// Raise chip select(s) so that displays can be individually configured
	digitalWrite(eye[0].tft_cs, HIGH);
	if (NUM_EYES > 1) digitalWrite(eye[1].tft_cs, HIGH);

	for (uint8_t e = 0; e < NUM_EYES; e++) {
		digitalWrite(eye[e].tft_cs, LOW);
		tft.setRotation(0);
		//tft.fillScreen(TFT_BLACK);

		digitalWrite(eye[e].tft_cs, HIGH);
	}

#if defined(DISPLAY_BACKLIGHT) && (DISPLAY_BACKLIGHT >= 0)
	Serial.println("Backlight now on!");
	analogWrite(DISPLAY_BACKLIGHT, BACKLIGHT_MAX);
#endif
	Serial.println("Eye Begins");
	startTime = millis(); // For frame-rate calculation

	xTaskCreatePinnedToCore(
		Task1code,   /* Task function. */
		"Task1",     /* name of task. */
		10000,       /* Stack size of task */
		NULL,        /* parameter of the task */
		1,           /* priority of the task */
		&Task1,      /* Task handle to keep track of created task */
		0);          /* pin task to core 1 */
	delay(500);

}

// MAIN LOOP -- runs continuously after setup() ----------------------------
int eyeCmdX = 0;
int eyeCmdY = 0;
int16_t eyeSpeed = 2;
int16_t xAccumulator = 64;
int16_t yAccumulator = 64;
int16_t xInc = 1;
int16_t yInc = 1;
int16_t irisCmd;
int16_t irisSpeed = 1;
bool Catseye = false;
bool evil = false;
bool blink = false;
int8_t blinkphase = 0;
int16_t blinktime = 0;
int16_t blinkspeed = 35;

bool EyeInPos()
{
	bool ret = (eyeCmdX == eye[0].xang && eyeCmdY == eye[0].yang);
	return ret;
}
bool IrisInPos()
{
	return (eye[0].irissize == irisCmd);
}

void MoveIris()
{
	if (eye[0].irissize > irisCmd)
	{
		eye[0].irissize -= irisSpeed;
		if (eye[0].irissize < irisCmd)
			eye[0].irissize = irisCmd;
	}
	if (eye[0].irissize < irisCmd)
	{
		eye[0].irissize += irisSpeed;
		if (eye[0].irissize > irisCmd)
			eye[0].irissize = irisCmd;
	}
}

#define cocenter 9870

void MoveEye()
{

	xAccumulator += xInc * eyeSpeed;  // Divide by 128 to get back to fixed-point
	yAccumulator += yInc * eyeSpeed;
	// eyeSpeed = 0 - 70
	/*
	Serial.print(eyeSpeed);
	Serial.print("Acc:");
	Serial.print(xAccumulator);
	Serial.print(",");
	Serial.print(yAccumulator);
	Serial.print(" Inc:");
	Serial.print(xInc);
	Serial.print(",");
	Serial.print(yInc);
	Serial.print(" Act:");
	Serial.print(eye[0].xang);
	Serial.print(",");
	Serial.print(eye[0].yang);
	Serial.print(" Cmd");
	Serial.print(eyeCmdX);
	Serial.print(",");
	Serial.print(eyeCmdY);
	Serial.println(",");
	*/

	if (eye[0].xang > eyeCmdX)
	{
		eye[0].xang -= xAccumulator >> 7;
		if (eye[0].xang < eyeCmdX)
			eye[0].xang = eyeCmdX;
	}
	else
		if (eye[0].xang < eyeCmdX)
		{
			eye[0].xang += xAccumulator >> 7;
			if (eye[0].xang > eyeCmdX)
				eye[0].xang = eyeCmdX;
		}

	if (eye[0].yang > eyeCmdY)
	{
		eye[0].yang -= yAccumulator >> 7;
		if (eye[0].yang < eyeCmdY)
			eye[0].yang = eyeCmdY;
	}
	else
		if (eye[0].yang < eyeCmdY)
		{
			eye[0].yang += yAccumulator >> 7;
			if (eye[0].yang > eyeCmdY)
				eye[0].yang = eyeCmdY;
		}
	xAccumulator %= 128;  // Keep the remainder for the next iteration
	yAccumulator %= 128;

}

void commandPos(int8_t xpos, int8_t ypos, int8_t speed)
{
	eyeCmdX = xpos; eyeCmdY = ypos, eyeSpeed = speed;
	//calc adders.

	int16_t curx = (eye[0].xang) << 7;
	int16_t cury = (eye[0].yang) << 7;
	int32_t targetX = eyeCmdX << 7;
	int32_t targetY = eyeCmdY << 7;
	int32_t dx = abs(eyeCmdX - eye[0].xang);
	int32_t dy = abs(eyeCmdY - eye[0].yang);
	int32_t steps = max(abs(dx), abs(dy));
	xInc = (abs(targetX - curx)) / steps;
	yInc = (abs(targetY - cury)) / steps;
	xAccumulator = yAccumulator = 0;
	/*
	Serial.println("---------------------------------------------------");
	Serial.println("---------------------------------------------------");
	Serial.print(steps);
	Serial.print("Acc:");
	Serial.print(xAccumulator);
	Serial.print(",");
	Serial.print(yAccumulator);
	Serial.print(" Inc:");
	Serial.print(xInc);
	Serial.print(",");
	Serial.print(yInc);
	Serial.print(" Act:");
	Serial.print(eye[0].xang);
	Serial.print(",");
	Serial.print(eye[0].yang);
	Serial.print(" Cmd");
	Serial.print(eyeCmdX);
	Serial.print(",");
	Serial.print(eyeCmdY);
	Serial.println(",");
	Serial.print(" delta");
	Serial.print(dx);
	Serial.print(",");
	Serial.print(dy);
	Serial.println(",");
	*/
};

void commandIris(int8_t size, int8_t speed)
{
	irisCmd = size % 40; irisSpeed = speed;
};

struct eyetest
{
	int16_t xpos;  int16_t ypos; int8_t speed; int8_t eyeSize; int8_t eyeSpeed; bool cateye; bool delay;
};

#define NTESTS 19
eyetest Tests[NTESTS] =
{
	{ 0, 80, 50, 3, 1 , 0, 0 },
	{ 0, -80, 50, 4, 1 , 0 ,0},
	{ 0, 0, 5, 4, 1, 0 ,0},
	{ 0, 30, 7, 15, 2, 0 ,0},
	{ 0, 0, 7, 3, 1, 0 ,0},
	{ 80, 0, 5, 3, 1, 0,0 },
	{ -80, 0, 5, 3, 1, 0 ,0},
	{ 0, 0, 6, 2, 1, 0 ,0},
	{ -6, 42, 6, 20, 4, 0 ,0},
	{ 30, 30, 2, 1, 1, 0 ,0},
	{ -30, -10, 2, 1, 1, 0 ,0},
	{  0, 0, 2, 3, 3, 0 ,0},
	{  0, -50, 8, 3, 3, 0 ,0},
	{  0, 50, 8, 3, 3, 0 ,0},
	{  0, 0, 8, 3, 3, 0 ,0},
	{  50, 50, 10, 3, 1, 0 ,0},
	{  -50, -50, 10, 18, 1, 0 ,0},
	{  50, 50, 10, 3, 1, 0 ,0},
	{  0, 0, 15, 18, 1, 0 ,0},
};


#define GC9A01A_SLPIN 0x10     ///< Enter Sleep Mode
#define GC9A01A_SLPOUT 0x11    ///< Sleep Out
#define uS_TO_S_FACTOR 1000000  /* Conversion factor for micro seconds to seconds */
#define TIME_TO_SLEEP  10        /* Time ESP32 will go to sleep (in seconds) */


void loop()
{
	static uint32_t millcnt = 0;
	static int lastcnt = 0;
	static int nTest = -1;
	uint32_t mills = millis();
	uint16_t curms = mills - lastcnt;
	lastcnt = mills;
	millcnt += curms;

	static bool tfton = true;

	if (millcnt > 10000 & 0)
	{
		tfton = !tfton;
		millcnt = 0;
		if (tfton)
		{
			digitalWrite(eye[0].tft_cs, LOW);
			tft.writecommand(GC9A01A_SLPOUT);
			digitalWrite(eye[0].tft_cs, HIGH);
		}
		else
		{
			digitalWrite(eye[0].tft_cs, LOW);
			tft.writecommand(GC9A01A_SLPIN);
			digitalWrite(eye[0].tft_cs, HIGH);
			//gotosleep
			esp_sleep_enable_timer_wakeup(TIME_TO_SLEEP * uS_TO_S_FACTOR);
			Serial.println("Going to sleep now");
			delay(100);
			Serial.flush();
			esp_deep_sleep_start();
		}

	}

	DoFrame = true;
	for (uint8_t e = 0; e < NUM_EYES; e++)
	{
		updateEye(eye[e].xang, eye[e].yang);
	}
	if (EyeInPos() && IrisInPos())
	{
		nTest = (nTest + 1) % NTESTS;
		eyetest *tst = &Tests[nTest];

		int posx = 60 - random(120);
		int posy = 60 - random(120);
		int speed = random(20) + 2;
		int iris = random(40);
		int irisspeed = random(5) + 2;
		commandPos(posx, posy, speed);
		commandIris(iris, irisspeed);
		if (random(10) > 8)
			evil = !evil;
		if (random(50) > 45 && blinktime == 0)
		{
			blink = true;
			blinkphase = 0;
		}
		//commandPos(tst->xpos, tst->ypos, tst->speed);
		//commandIris(tst->eyeSize, tst->eyeSpeed);
		//Catseye = tst->cateye;
		//if( tst->delay)
		//   delay(3000);
	}
	if (!IrisInPos())
		MoveIris();
	if (!EyeInPos())
		MoveEye();


	//draw the frame
	while (DoFrame)
	{
		yield();
	};//all drawing is finished
	frame(0);

}

void Task1code(void * parameter)
{
	static int loop = 0;

	while (1)
	{
		if (DoFrame)
		{
			updateEye2(eye[0].xang, eye[0].yang);
			DoFrame = false;
		}

		vTaskDelay(1);
		yield();
	}
}